"""
Parse all the test Mapfiles in msautotests, write them to a new file,
and then test that these also parse correctly
"""
import os, logging, glob, json, shutil

from mappyfile.pprint import PrettyPrinter
from mappyfile.parser import Parser
from mappyfile.transformer import MapfileToDict
import pprint
import mappyfile

def create_copy(msautotest_fld, msautotest_copy):

    # first make a backup copy of msautotest


    if os.path.isdir(msautotest_copy):
        shutil.rmtree(msautotest_copy)
        logging.info("Removing %s...", msautotest_copy)

    logging.info("Copying %s to %s...", msautotest_fld, msautotest_copy)
    shutil.copytree(msautotest_fld, msautotest_copy)
    logging.info("Copying complete!")

    return msautotest_copy

def parse_mapfile(parser, transformer, pp, fn):
    logging.debug("Parsing %s", fn)

    try:
        ast = parser.parse_file(fn)
        #print(ast)
    except Exception as ex:
        logging.warning("%s could not be successfully parsed", fn)
        logging.exception(ex)
        raise

    try:
        d = transformer.transform(ast)
    except Exception as ex:
        logging.warning("%s could not be successfully transformed", fn)
        logging.exception(ex)
        raise
        
    #map_string = pp.pprint(d)
    #print(map_string)

    return d

def main(msautotest_fld, create_new_copy=True):

    msautotest_copy = os.path.join(os.path.dirname(msautotest_fld), "msautotest_mappyfile")

    if create_new_copy:
        create_copy(msautotest_fld, msautotest_copy)

    parser = Parser()
    transformer = MapfileToDict()
    pp = PrettyPrinter()

    # these two maps aren't in utf8, see https://github.com/mapserver/mapserver/pull/5460
    #ignore_list = ["wms_inspire_scenario1.map","wms_inspire_scenario2.map"] 
    ignore_list = []

    mapfiles = glob.glob(msautotest_fld + '/**/*.map')
    mapfiles = [f for f in mapfiles if os.path.basename(f) not in ignore_list]
    
    for fn in mapfiles:

        d = parse_mapfile(parser, transformer, pp, fn)
        output_file = fn.replace(msautotest_fld, msautotest_copy)
        mf = mappyfile.utils.write(d, output_file)

        # now try reading it again
        d = parse_mapfile(parser, transformer, pp, output_file)

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    fld = r"D:\GitHub\mapserver\msautotest"
    main(fld, True)
    print("Done!")


import os, logging
from io import open
from lark import Lark, ParseError

class Parser(object):

    def __init__(self, cwd="", expand_includes=True):
        self.cwd = cwd
        self.expand_includes = expand_includes
        self.g = self.load_grammar("mapfile.g")

        
    def load_grammar(self, grammar_file):

        gf = os.path.join(os.path.dirname(__file__), grammar_file)
        grammar_text = open(gf).read()

        return Lark(grammar_text, parser='earley', lexer='standard')

    def strip_quotes(self, s):
        return s.strip("'").strip('"')

    def load_includes(self, text):

        lines = text.split('\n')
        includes = {}

        for idx, l in enumerate(lines):
            if l.strip().lower().startswith('include'):
                if '#' in l:
                    l = l[:l.index('#')]

                parts = [p for p in l.split()]

                assert (len(parts) == 2)
                assert (parts[0].lower() == 'include')
                fn = os.path.join(self.cwd, self.strip_quotes(parts[1]))
                try:
                    include_text = self.open_file(fn)
                except IOError as ex:
                    logging.warning("Include file '%s' not found", fn)
                    raise ex
                # recursively load any further includes
                includes[idx] = self.load_includes(include_text)
    
        for idx, txt in includes.items():
            lines.pop(idx) # remove the original include
            lines.insert(idx, txt)

        return '\n'.join(lines)

    def open_file(self, fn):
        try:
            return open(fn, "r", encoding="utf-8").read() # specify Unicode for Python 2.7
        except UnicodeDecodeError as ex:
            logging.debug(ex)
            logging.error("Please check the encoding for %s. All Mapfiles should be in utf-8 format. ", fn)
            raise

    def parse_file(self, fn):

        self.cwd = os.path.dirname(fn)

        text = self.open_file(fn)
        return self.parse(text)

    def parse(self, text):

        if self.expand_includes == True:
            text = self.load_includes(text)

        text += '\n\n'

        try:
            return self.g.parse(text)
        except ParseError:
            # try with the grammar that ignores new lines
            # this is 3x slower than with new lines
            self.g = self.load_grammar("mapfile.nobreaks.g")
            return self.g.parse(text)




from mappyfile.parser import Parser
from mappyfile.transformer import MapfileToDict
from mappyfile.pprint import PrettyPrinter
import codecs

def load(fn, cwd=None):

    p = Parser(cwd=cwd)
    ast = p.parse_file(fn)
    m = MapfileToDict()
    d = m.transform(ast)

    return d    

def loads(s, cwd="", expand_includes=True):
    p = Parser(cwd=cwd, expand_includes=expand_includes)
    ast = p.parse(s)
    m = MapfileToDict()
    d = m.transform(ast)

    return d   

def write(d, output_file, indent=4):

    map_string = _pprint(d, indent)
    _save(output_file, map_string)

    return output_file

def dumps(d):
    return _pprint(d)

def find(lst, key, value):
    """
    When looking for an item by value also check for the value 
    surrounded by apostrophes
    """
    obj = __find__(lst, key, value)

    if not obj:
        v = "'%s'" % value
        obj = __find__(lst, key, v)

    if not obj:
        v = '"%s"' % value
        obj = __find__(lst, key, v)

    return obj

def __find__(lst, key, value):
    return next((item for item in lst if item[key.lower()] == value), None)

def findall(lst, key, value):
    possible_values = ("'%s'" % value, '"%s"' % value)

    return (item for item in lst if item[key.lower()] in possible_values)

def _save(output_file, map_string):

    with codecs.open(output_file, "w", encoding="utf-8") as f:
        f.write(map_string)

def _pprint(d, indent=4):
    pp = PrettyPrinter(indent=indent)
    return pp.pprint(d)



# -*- coding: utf-8 -*-
import pytest
from mappyfile.pprint import PrettyPrinter
import mappyfile

def test_nested_quotes():
    """
    If values contain quotes then make sure they are escaped
    shp2img -m C:\Temp\msautotest\misc\ogr_vrtconnect.tmp.map    
    """
    s = """
    LAYER
        NAME shppoly
        TYPE polygon
        CONNECTIONTYPE OGR
        CONNECTION '<OGRVRTDataSource><OGRVRTLayer name="poly"><SrcDataSource relativeToVRT="0">data/shppoly</SrcDataSource><SrcLayer>poly</SrcLayer></OGRVRTLayer></OGRVRTDataSource>'
    END"""

    ast = mappyfile.loads(s)
    pp = PrettyPrinter(indent=0, quote='"', newlinechar=" ") # expected
    res = pp.pprint(ast)
    exp = r'LAYER NAME shppoly TYPE polygon CONNECTIONTYPE OGR CONNECTION "<OGRVRTDataSource><OGRVRTLayer name=\"poly\">' \
        r'<SrcDataSource relativeToVRT=\"0\">data/shppoly</SrcDataSource><SrcLayer>poly</SrcLayer></OGRVRTLayer></OGRVRTDataSource>" END'
    assert(res == exp)

def test_standardise_quotes():

    v = '"the_geom from (select * from road where "name_e"=\'Trans-Canada Highway\' order by gid) as foo using unique gid using srid=3978"'

    pp = PrettyPrinter(indent=0, quote='"', newlinechar=" ") # expected
    v2 = pp.standardise_quotes(v)
    exp = r'''"the_geom from (select * from road where \"name_e\"='Trans-Canada Highway' order by gid) as foo using unique gid using srid=3978"'''
    assert(v2 == exp)

    pp = PrettyPrinter(indent=0, quote="'", newlinechar=" ") # expected
    v2 = pp.standardise_quotes(v)
    exp = r"""'the_geom from (select * from road where "name_e"=\'Trans-Canada Highway\' order by gid) as foo using unique gid using srid=3978'"""
    assert(v2 == exp)

def test_double_attributes():

    s = 'MAP CONFIG "MS_ERRORFILE" "stderr" END'
    ast = mappyfile.loads(s)
    pp = PrettyPrinter(indent=0, quote='"', newlinechar=" ")
    res = pp.pprint(ast)
    assert(res == 'MAP CONFIG "MS_ERRORFILE" "stderr" END')

    s = 'MAP CONFIG "MS_ERRORFILE" "stderr" END'
    ast = mappyfile.loads(s)
    pp = PrettyPrinter(indent=0, quote="'", newlinechar=" ") # expected
    res = pp.pprint(ast)
    assert(res == "MAP CONFIG 'MS_ERRORFILE' 'stderr' END")
        
def test_already_escaped():
    """
    Don't escape an already escaped quote
    """
    s = r'CLASS EXPRESSION "\"Tignish" END'
    ast = mappyfile.loads(s)
    pp = PrettyPrinter(indent=0, quote='"', newlinechar=" ")
    res = pp.pprint(ast)
    exp = r'CLASS EXPRESSION "\"Tignish" END'
    assert(res == exp)

    s = r"CLASS EXPRESSION '\'Tignish' END"
    ast = mappyfile.loads(s)
    pp = PrettyPrinter(indent=0, quote="'", newlinechar=" ")
    res = pp.pprint(ast)
    exp = r"CLASS EXPRESSION '\'Tignish' END"
    assert(res == exp)

def test_format_list():

    s = """
        CLASS
            STYLE
                COLOR 173 216 230
            END
            STYLE
                OUTLINECOLOR 2 2 2
                WIDTH 1
                LINECAP BUTT
                PATTERN
                    5 5
                    10 10
                END
            END
        END
    """

    ast = mappyfile.loads(s)
    pp = PrettyPrinter(indent=0) # expected

    k = "pattern"
    lst = [[5, 5, 10, 10]]

    assert(pp.is_paired_list(k))
    r = pp.process_list(k, lst, 0)
    exp = [u'PATTERN', '5 5\n10 10', u'END']
    assert(r == exp)

def test_unicode():

    s = u"""
    MAP
        METADATA
          "ows_title" "éúáí"
        END
    END
    """
    ast = mappyfile.loads(s)
    pp = PrettyPrinter(indent=0, quote="'", newlinechar=" ")
    res = pp.pprint(ast)
    exp = u"MAP METADATA 'ows_title' 'éúáí' END END"
    assert(res == exp)

def test_config():

    s = """
    MAP
        config "MS_ERRORFILE" "my.log"
    END
    """
    ast = mappyfile.loads(s)
    pp = PrettyPrinter(indent=0, quote="'", newlinechar=" ")
    res = pp.pprint(ast)
    #print(res)
    exp = u"MAP CONFIG 'MS_ERRORFILE' 'my.log' END"
    assert(res == exp)

def run_tests():        
    #pytest.main(["tests/test_pprint.py::test_format_list"])
    pytest.main(["tests/test_pprint.py"])

if __name__ == "__main__":
    run_tests()
    #test_double_attributes()
    print("Done!")



import sys
is_python3 = sys.version_info.major == 3
if is_python3:
    unicode = str
    
from mappyfile.tokens import COMPOSITE_NAMES, ATTRIBUTE_NAMES, SINGLETON_COMPOSITE_NAMES, REPEATED_KEYS

ALL_KEYWORDS = COMPOSITE_NAMES.union(ATTRIBUTE_NAMES).union(SINGLETON_COMPOSITE_NAMES)


class PrettyPrinter(object):
    def __init__(self, indent=4, spacer=" ", quote='"', newlinechar="\n"):
        """
        Option use "\t" for spacer with an indent of 1
        """

        assert (quote == "'" or quote == '"')

        self.indent = indent
        self.spacer = spacer * self.indent
        self.newlinechar = newlinechar
        self.quote = quote
        self.end = u"END"

        if self.quote == "'":
            self.altquote = '"'
        else:
            self.altquote = "'"

    def whitespace(self, level, indent):

        return self.spacer * (level + indent)

    def singular(self, s):
        if s == 'points':
            return s
        elif s.endswith('es'):
            return s[:-2]
        return s[:-1]

    def is_paired_list(self, key):
        """
        Temporary workaround for POINT and PATTERNS which currently must have
        pairs on the same line
        """
        if key in ('pattern','points'):
            return True
        else:
            return False

    def is_block_list(self, lst):
        if len(lst) == 1 and isinstance(lst[0], list):
            return True
        else:
            return False

    def is_list_of_lists(self, lst):
        return all(isinstance(l, list) for l in lst)

    def process_dict(self, key, d, level):

        lines = []
        lines.append(self.whitespace(level, 1) + self.format_key(key))

        for k, v in d.items():
            lines.append(self.format_line(self.whitespace(level, 2), k, v))

        lines.append(self.whitespace(level, 1) + self.end)

        return lines

    def process_list(self, key, lst, level):
        
        lines = []
        spacer = self.spacer * (level + 1)
        block_list = self.is_block_list(lst)

        # add the items in the list

        if block_list:
            # add KEYWORD to signify start of block
            lines.append(spacer + self.format_key(key))
            list_spacer = self.spacer * (level + 2)
            s = self.format_list(key, lst[0], list_spacer)
            lines.append(s)
        else:
            # put all parts on same line
            if key in REPEATED_KEYS:
                for v in lst:
                    lines.append(self.format_line(spacer, key, v))
            else:
                s = " ".join(map(self.format_value, lst))
                s = self.format_line(spacer, key, s)
                lines.append(s)
       
        if block_list:
            # add END to close the block
            s = self.spacer * (level + 1) + self.end
            lines.append(s)

        return lines


    def format_list(self, key, val, spacer):
        """
        Print out a list of values
        """
        vals = list(map(str, val))

        if self.is_paired_list(key):
            # join the values together so each line has a pair
            vals = zip(vals[::2], vals[1::2])
            vals = ["%s %s" % (v[0], v[1]) for v in vals]
        
        s = self.newlinechar.join([spacer + v for v in vals])

        return s

    def format_attribute(self, val):
        if self.in_quotes(val):
            val = self.standardise_quotes(val)
        else:
            val = self.add_quotes(val)

        return self.format_value(val)

    def format_value(self, val):
        """
        Print out a string
        """

        if isinstance(val, (unicode, str)):
            val = self.standardise_quotes(val)
        
        try:
            val = unicode(val)
        except UnicodeDecodeError:
            # obj is byte string
            ascii_text = str(val).encode('string_escape')
            val = unicode(ascii_text)

        return val

    def add_quotes(self, val):
        return "%s%s%s" % (self.quote, val, self.quote)

    def in_quotes(self, val):

        if (val.startswith(self.quote) and val.endswith(self.quote)) \
            or  (val.startswith(self.altquote) and val.endswith(self.altquote)):
            return True
        else:
            return False

    def escape_quotes(self, val):
        """
        Escape any quotes in a value
        """
        if val.startswith(self.quote) and val.endswith(self.quote):
            # make sure any previously escaped quotes are not re-escaped
            middle = val[1:-1].replace("\\" + self.quote, self.quote)
            middle = middle.replace(self.quote, "\\" + self.quote)
            val = "%s%s%s" % (self.quote, middle, self.quote)

        return val

    def standardise_quotes(self, val):
        """
        Change the quotes used to wrap a value to the pprint default
        E.g. "val" to 'val' or 'val' to "val"
        """
        if val.startswith(self.altquote) and val.endswith(self.altquote):
            middle = val[1:-1]
            val = "%s%s%s" % (self.quote, middle, self.quote)

        val = self.escape_quotes(val)

        return val

    def format_key(self, key):
        """
        All non-quoted keys should be in uppercase
        Otherwise do not modify input
        """
        if self.in_quotes(key):
            key = self.format_value(key) 
        else:
            if key.lower() in ALL_KEYWORDS:
                key = self.format_value(key.upper())
            else:
                key = self.format_value(self.add_quotes(key.lower()))

        return key

    def format_line(self, spacer, key, value):
        return self.__format_line(spacer, self.format_key(key), self.format_value(value))

    def __format_line(self, spacer, key, value):

        tmpl = u"{spacer}{key} {value}"
        d = {
            "spacer": spacer,
            "key": key,
            "value": value
            }
        return tmpl.format(**d)

    def is_composite(self, val):

        if isinstance(val, dict) and "__type__" in val.keys():
            return True
        else:
            return False

    def is_hidden_container(self, key, val):
        """
        The key is not one of the Mapfile keywords, and its 
        values are a list
        """

        if key not in ALL_KEYWORDS and isinstance(val, list):
            return True
        else:
            return False

    def pprint(self, composites):
        """
        Print out a nicely indented Mapfile
        """

        lines = []

        # if only a single composite is used then cast to list
        if not isinstance(composites, list):
            composites = [composites]

        # allow for multiple root composites
        for composite in composites:
            lines += self._format(composite)

        return self.newlinechar.join(lines)

    def _format(self, composite, level=0):
        """
        TODO Refactor this and create functions for each unique type
        """
        lines = []
        if isinstance(composite, dict) and  '__type__' in composite.keys():
            type_ = composite['__type__']
            assert(type_ in COMPOSITE_NAMES.union(SINGLETON_COMPOSITE_NAMES))
            is_hidden = False
            s = self.whitespace(level, 0) + type_ .upper()                          
            lines.append(s)
            # we can now filter out the type property so the rest of the 
            # values are displayed
            items = ((k, v) for k, v in composite.items() if k != '__type__')
        else:
            # hidden container
            assert(len(composite.keys()) == 1)
            is_hidden = True
            items = enumerate(composite.values()[0])

        for key, value in items:
       
            if self.is_hidden_container(key, value): # HiddenContainer
                # now recursively print all the items in the container
                if self.is_block_list(value):
                    k = self.singular(key)
                    lines += self.process_list(k, value, level)
                else:
                    for v in value:
                        lines += self._format(v, level + 1)

            elif self.is_composite(value): # Container
                lines += self._format(value, level + 1)
            else:
                if key in SINGLETON_COMPOSITE_NAMES:                    
                    lines += self.process_dict(key, value, level)
                elif isinstance(value, dict): 

                    if key == "config":
                        # config declaration allows for pairs of values
                        value = ["%s %s" % (self.format_key(k), self.format_attribute(v)) for k,v in value.items()]
                    key = self.format_key(key) # format the "parent" key
                    for v in value:
                        # keys and values are already formatted so do not format them again                 
                        lines.append(self.__format_line(self.whitespace(level, 1), key, v))
                elif isinstance(value, list):                    
                    if self.is_list_of_lists(value):
                        # value is list of lists, so create composite type for each list e.g. several POINTS in a FEATURE
                        for l in value:
                            lines += self.process_list(key, [l], level)
                    else:
                        lines += self.process_list(key, value, level)
                else:
                    comp_type = composite.get("__type__", "")                   
                    if comp_type == "metadata":
                        # don't add quotes to key or value, but standardise them if present
                        key = self.standardise_quotes(key)
                        value = self.standardise_quotes(value)
                        lines.append(self.__format_line(self.whitespace(level, 1), key, value))
                    else:
                        lines.append(self.format_line(self.whitespace(level, 1), key, value))

        if not is_hidden: # Container
            # close the container block with an END            
            s = self.whitespace(level, 0) + self.end
            lines.append(s)

        return lines



import os, logging
import pytest
import mappyfile
from mappyfile.parser import Parser
from mappyfile.transformer import MapfileToDict

def test_all_maps():

    sample_dir = os.path.join(os.path.dirname(__file__), "sample_maps")

    p = Parser(expand_includes=False)

    for fn in os.listdir(sample_dir):
        print(fn)
        try:
            ast = p.parse_file(os.path.join(sample_dir, fn))
            #ast.to_png_with_pydot(r'C:\Temp\Trees\%s.png' % os.path.basename(fn))
        except:
            logging.warning("Cannot process %s ", fn)
            raise

def test_includes():
    p = Parser()
   
    ast = p.parse_file('./tests/samples/include1.map')
    m = MapfileToDict()

    d = (m.transform(ast)) # works
    print(mappyfile.dumps(d))

def run_tests():        
    pytest.main(["tests/test_sample_maps.py"])

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    test_all_maps()
    #run_tests()
    print("Done!")


import pytest

from mappyfile.ordereddict import DefaultOrderedDict

def test_dict():

    d = DefaultOrderedDict()
    d["key"] = ["value"]

    print(d["key"])

def run_tests():        
    #pytest.main(["tests/test_ordereddict.py::test_dict"])
    pytest.main(["tests/test_ordereddict.py"])

if __name__ == '__main__':
    run_tests()
    #test_dict()


"""
Module to transform an AST (Abstract Syntax Tree) to a 
Python dict structure
"""

from collections import OrderedDict
from lark import Transformer, Tree
from mappyfile.tokens import ATTRIBUTE_NAMES, COMPOSITE_NAMES, SINGLETON_COMPOSITE_NAMES, REPEATED_KEYS
from mappyfile.ordereddict import DefaultOrderedDict

def plural(s):

    if s == 'points':
        return s
    elif s.endswith('s'):
        return s +'es'
    else:
        return s +'s'

def dict_from_tail(t):
    """
    VALIDATION blocks can also have attributes such as qstring
    Values then have 3 parts - [('attr', u'qstring', u"'.'")]
    METADATA blocks have a simple 2 part form - [[u'"ows_enable_request"', u'"*"']]
    """
    d = OrderedDict()

    for v in t:
        if len(v) == 2:
            d[v[0]] = v[1]
        elif len(v) == 3:
            d[v[1]] = v[2]
        else:
            raise ValueError("Unsupported block '%s'", str(v))
    return d

class MapfileToDict(Transformer):
  
    def start(self, children):
        """
        Not limited to MAP..END parsing of partial files or composites is 
        also possible
        """

        composites = []

        for child in children:
            assert(child[0] == 'composite')
            composites.append(child[2])

        # only return a list when there are multiple root composites (e.g. several CLASSes)
        if len(composites) == 1:
            return composites[0]
        else:
            return composites

    def repeated_key(self, d, k, v):
        """
        Allow the key to be added multiple times to its parent
        """
        if k not in d.keys():
            d[k] = [v]
        else:
            d[k].append(v)

        return d

    def composite(self, t):
        """
        Handle the composite types e.g. CLASS..END
        [Tree(composite_type, [Token(__CLASS8, 'CLASS')]), Tree(composite_body, [('attr', u'name', "'test'")])]
        """
        if len(t) == 3:
            # Parser artifact. See LINE-BREAK FLUIDITY in parsing_decisions.txt
            type_, attr, body = t
        elif len(t) == 2:
            type_, body = t
            attr = None
        else:
            assert(len(t) == 1)
            type_, attr, body = t[0]
            #print type_, attr, body
            if attr in ("metadata","validation"):
                body['__type__'] = attr
                return ('composite', attr, body)

        if isinstance(body, tuple):
            assert body[0] == 'attr' or body[1] == 'points' or body[1] == 'pattern', body  # Parser artefacts
            body = [body]
        else:
            body = body.children

        type_ = type_.children[0].lower()

        assert type_ in COMPOSITE_NAMES.union(SINGLETON_COMPOSITE_NAMES)

        if attr:
            body = [attr] + body

        for x in body:
            assert isinstance(x, tuple), x

        composites = DefaultOrderedDict(list)
        d = DefaultOrderedDict(OrderedDict)

        for itemtype, k, v in body:          

            if itemtype == 'attr':

                # TODO tidy-up the code below
                if k in REPEATED_KEYS:
                    d = self.repeated_key(d, k, v)
                elif k == 'config':
                    # CONFIG can be repeated, but with pairs of strings as a value
                    if 'config' not in d.keys():
                        d[k] = OrderedDict()
                    d[k][v[0]] = v[1]
                else:
                    d[k] = v

            elif itemtype == 'composite' and k in SINGLETON_COMPOSITE_NAMES:
                # there can only ever be one instance of these
                composites[k] = v # defaultdict using list
            elif itemtype == 'composite':
                composites[k].append(v)
            else:
                raise ValueError("Itemtype '%s' unknown", itemtype)
        
        for k, v in composites.items(): # collection of all items e.g. at the map level this is status, metadata etc. 

            if k in SINGLETON_COMPOSITE_NAMES:
                d[k] = v
            else:
                d[plural(k)] = v
                
        d['__type__'] = type_
        return ('composite', type_, d)

    def attr(self, children):
        name = children[0].children[0]
        
        if isinstance(name, Tree):
            name = name.children[0] # Solve a parser artefact for composite names

        name = name.lower()
        # VALIDATION and METADATA blocks can take any unquoted string as a key
        #assert name in ATTRIBUTE_NAMES, name 

        value = children[1:]
        if len(value) == 1:
            value = value[0]
        return 'attr', name, value

    def projection(self, t):
        return ('composite', 'projection', t)

    def metadata(self, t):
        """
        Create a dict for the metadata items
        """
        d = dict_from_tail(t)
        return ('composite', 'metadata', d)

    def points(self, t):
        return ('composite', 'points', t)

    def pattern(self, t):
        # http://www.mapserver.org/mapfile/style.html
        return ('composite', 'pattern', t)

    def values(self, t):
        d = dict_from_tail(t)
        return ('composite', 'values', d)

    def validation(self, t):
        """
        Create a dict for the validation items
        """
        d = dict_from_tail(t)
        return ('composite', 'validation', d)

    # for expressions

    def comparison(self, t):
        parts = [str(p) for p in list(t)]
        v = " ".join(parts)
        return "( %s )" % v
    def and_test(self, t):
        v = " and ".join(t)
        return "( %s )" % v
    def or_test(self, t):
        v = " or ".join(t)
        return "( %s )" % v
    def compare_op(self, t):
        v = t[0]
        return v

    def not_expression(self, t):
        return "not %s" % t[0]
    def expression(self, t):
        return "(%s)" % t[0]
    def add(self, t):
        return "%s + %s" % tuple(t)

    def runtime_var(self, t):
        v = t[0]
        return v

    def regexp(self, t):
        """
        E.g. regexp(u'/^[0-9]*$/')
        """
        v = t[0]
        return v

    # for functions

    def func_call(self, t):
        """
        For function calls e.g. TEXT (tostring([area],"%.2f"))
        """
        func, params = t
        func_name = func.children[0]
        v = "(%s(%s))" % (func_name, params)
        return v

    def func_params(self, t):
        params = ",".join(str(v) for v in t)
        return params

    def attr_bind(self, t):
        v = t[0]
        return "[%s]" % v

    # basic types

    def int(self, t):
        v = t[0]
        return int(v)
    def float(self, t):
        v = t[0]
        return float(v)
    def bare_string(self, t):
        if t:
            v = t[0]
        else:
            v = t            
        return v
    def string(self, t):
        v = t[0].value
        return v
    def path(self, t):
        v = t[0]
        return v
    def string_pair(self, t):
        a, b = t
        return [a, b]
    def int_pair(self, t):
        a, b = t
        return [a, b]
    def list(self, t):
        # http://www.mapserver.org/mapfile/expressions.html#list-expressions
        return "{%s}" % ",".join([str(v) for v in t])






import re
from setuptools import setup

__version__,= re.findall('__version__ = "(.*)"', open('mappyfile/__init__.py').read())

def readme():
    with open('README.rst') as f:
        return f.read()

setup(name='mappyfile',
      version=__version__,
      description='A pure Python MapFile parser for working with MapServer',
      long_description=readme(),
      classifiers=[
        # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Intended Audience :: Developers',
        'Topic :: Text Processing :: Linguistic',
        'Topic :: Software Development :: Build Tools'
      ],
      package_data = {
        '': ['*.g']
      },
      url='http://github.com/geographika/mappyfile',
      author='Seth Girvin',
      author_email='sethg@geographika.co.uk',
      license='MIT',
      packages=['mappyfile'],
      install_requires=['lark-parser'],
      zip_safe=False)


import pytest
from mappyfile.parser import Parser
from mappyfile.transformer import MapfileToDict

def test_processing_directive():

    s = """
    LAYER
        NAME 'ProcessingLayer'
        PROCESSING 'BANDS=1'
        PROCESSING 'CONTOUR_ITEM=elevation'
        PROCESSING 'CONTOUR_INTERVAL=20'
    END
    """

    p = Parser()
    ast = p.parse(s)
    t = MapfileToDict()
    d = t.transform(ast)
    assert(len(d["processing"]) == 3)

def test_config_directive():

    s = """
    MAP
        NAME 'ConfigMap'
        CONFIG MS_ERRORFILE "stderr"
        CONFIG "PROJ_DEBUG" "OFF"
        CONFIG "ON_MISSING_DATA" "IGNORE"
    END
    """

    p = Parser()
    ast = p.parse(s)
    t = MapfileToDict()
    d = t.transform(ast)
    assert(len(d["config"]) == 3)

def test_metadata():

    s = """
    MAP
        METADATA
            'wms_enable_request'  '*'     
        END
    END
    """

    p = Parser()
    ast = p.parse(s)
    t = MapfileToDict()
    d = t.transform(ast)
    #print(dict(d["metadata"]))
    assert(d["metadata"]["'wms_enable_request'"] == "'*'")

def run_tests():        
    #pytest.main(["tests/test_transformer.py::test_config_directive"])
    pytest.main(["tests/test_transformer.py"])

if __name__ == '__main__':
    #run_tests()
    test_metadata()
    print("Done!")


import os, logging
import pytest
from mappyfile.parser import Parser
from mappyfile.pprint import PrettyPrinter
import mappyfile
from mappyfile.transformer import MapfileToDict

def output(s):
    """
    Parse, transform, and pretty print 
    the result
    """
    p = Parser()
    m = MapfileToDict()
    
    ast = p.parse(s)
    #print(ast)
    d = m.transform(ast)
    #print(d)
    pp = PrettyPrinter(indent=0, newlinechar=" ", quote="'")
    return pp.pprint(d)


def check_result(s):
    try:
        s2 = output(s)
        assert(s == s2)
    except AssertionError:
        logging.info(s)
        logging.info(s2)
        raise


def test_layer():
    s = "LAYER NAME 'Test' END"
    check_result(s)


def test_class():

    s = "CLASS NAME 'Test' END"
    check_result(s)


def test_style():

    s = "STYLE COLOR 0 0 255 WIDTH 5 LINECAP BUTT END"
    check_result(s)


def test_style_pattern():

    s = """
    STYLE
        PATTERN 5 5 END 
    END
    """

    exp = "STYLE PATTERN 5 5 END END"
    assert(output(s) == exp)


def test_style_pattern2():

    s = """
    STYLE 
        PATTERN 
            5 5
        END 
    END
    """

    exp = "STYLE PATTERN 5 5 END END"
    assert(output(s) == exp)


def test_style_pattern3():
    """
    Test a STYLE on one line can be parsed
    """
    s = "STYLE PATTERN 5 5 END END"
    exp = "STYLE PATTERN 5 5 END END"
    assert(output(s) == exp)


def test_style_pattern4():
    """
    Test pattern values on separate lines are valid
    """
    s = """
    STYLE 
        PATTERN 
            5
            5 
        END 
    END
    """
    exp = "STYLE PATTERN 5 5 END END"
    assert(output(s) == exp)

@pytest.mark.xfail
def test_style_pattern5():
    """
    Test pattern with odd number of values
    This should fail with the following error message
    UnexpectedToken: Unexpected token Token(_END, 'END') at line 3, column 27.
    """
    s = """
    STYLE 
        PATTERN 6 4 2 4 6 END
    END
    """
    exp = "STYLE PATTERN 6 4 2 4 6 END END"
    assert(output(s) == exp)

def test_metadata():
    """
    Parse metadata directly
    """
    s = """
    METADATA 
        'wms_title' 'Test simple wms'
    END
    """
    exp = """METADATA 'wms_title' 'Test simple wms' END"""
    assert(output(s) == exp)

def test_metadata_unquoted():
    """
    The METADATA block doesn't need quotes 
    (as long as values don't have spaces)
    """
    s = """
    METADATA 
        wms_title my_title
    END
    """
    exp = """METADATA wms_title my_title END"""
    #print output(s)
    assert(output(s) == exp)

def test_validation():
    """
    Parse validation block directly
    """
    s = """
    VALIDATION
        "field1" "^[0-9,]+$"
        "field2" "-1"
    END
    """
    #print output(s)
    exp = """VALIDATION 'field1' '^[0-9,]+$' 'field2' '-1' END"""
    assert(output(s) == exp)

def test_layer_text_query():
    s = """
    CLASS
        TEXT (tostring([area],"%.2f"))
    END
    """
    exp = """CLASS TEXT ((tostring([area],"%.2f"))) END"""
    assert(output(s) == exp)


def test_label():
    s = """
    LABEL
      COLOR  0 0 0
      FONT Vera
      TYPE truetype
      SIZE 8
      POSITION AUTO
      PARTIALS FALSE
      OUTLINECOLOR 255 255 255			
    END 
    """
    exp = "LABEL COLOR 0 0 0 FONT Vera TYPE truetype SIZE 8 POSITION AUTO PARTIALS FALSE OUTLINECOLOR 255 255 255 END"
    assert(output(s) == exp)


def test_output_format():

    s = """
    MAP
        IMAGETYPE grid
        OUTPUTFORMAT
          NAME grid2
          DRIVER "GDAL/AAIGRID"
          IMAGEMODE INT16
          FORMATOPTION 'NULLVALUE=-99'
        END
    END
    """
    exp = "MAP IMAGETYPE grid OUTPUTFORMAT NAME grid2 DRIVER 'GDAL/AAIGRID' IMAGEMODE INT16 FORMATOPTION 'NULLVALUE=-99' END END"
    assert(output(s) == exp)


def test_class_symbol():
    s = """
    CLASS
        STYLE # a shadow
            COLOR 151 151 151
            SYMBOL [symbol]
            OFFSET 2 2
            SIZE [size]
        END
    END
    """
    exp = "CLASS STYLE COLOR 151 151 151 SYMBOL [symbol] OFFSET 2 2 SIZE [size] END END"
    assert(output(s) == exp)


def test_filter():
    s = """
    LAYER
        NAME 'filters_test002'
        FILTER 'aitkin'i
    END
    """

    exp = "LAYER NAME 'filters_test002' FILTER 'aitkin'i END"
    assert(output(s) == exp)


def test_regex():
    s = r"""
    LAYER
        NAME 'regexp-example'
        FILTERITEM 'placename'
        FILTER \hotel\
    END
    """
    exp = r"LAYER NAME 'regexp-example' FILTERITEM 'placename' FILTER \hotel\ END"
    assert(output(s) == exp)


def test_feature():
    """
    With multiple points
    """

    s = """
        LAYER        
            FEATURE
                POINTS
                    0 10
                END
                POINTS
                    -20 20
                    20 20
                    -20 -20
                    0 -30
                    20 -20
                    -20 20
                    -20 20
                    30 30
                END
            END
        END
    """

    exp = "LAYER FEATURE POINTS 0 10 END POINTS -20 20 20 20 -20 -20 0 -30 20 -20 -20 20 -20 20 30 30 END END END"
    assert(output(s) == exp)


def test_symbol():

    s = '''
    SYMBOL
        NAME 'triangle'
        TYPE VECTOR
        POINTS
            0 4
            2 0
            4 4
            0 4
        END
        FILLED TRUE
    END
    '''
    exp = "SYMBOL NAME 'triangle' TYPE VECTOR FILLED TRUE POINTS 0 4 2 0 4 4 0 4 END END"
    assert(output(s) == exp)


def test_class_expression1():
    s = '''
    CLASS
      TEXT ([area])
    END
    '''
    exp = "CLASS TEXT ([area]) END"
    assert(output(s) == exp)


def test_class_expression2():
    """
    shp2img -m C:\Temp\msautotest\query\text.tmp.map  -l text_test002 -o c:\temp\tmp_onl0lk.png
    """
    s = '''
    CLASS
      TEXT ("[area]")
    END
    '''
    exp = 'CLASS TEXT ("[area]") END'
    assert(output(s) == exp)


def test_complex_class_expression():
    s = '''
    CLASS
      TEXT ("Area is: " + tostring([area],"%.2f"))
    END
    '''         
    exp = '''CLASS TEXT ("Area is: " + (tostring([area],"%.2f"))) END'''
    assert(output(s) == exp)


def test_or_expressions():
    """
    See http://www.mapserver.org/mapfile/expressions.html#expressions
    """

    s = '''
    CLASS
        EXPRESSION ("[style_class]" = "10" OR "[style_class]" = "20")
    END
    '''

    exp = 'CLASS EXPRESSION (( ( "[style_class]" = "10" ) or ( "[style_class]" = "20" ) )) END'
    assert(output(s) == exp)

    s = '''
    CLASS
        EXPRESSION ("[style_class]" = "10" || "[style_class]" = "20")
    END
    '''

    exp = 'CLASS EXPRESSION (( ( "[style_class]" = "10" ) or ( "[style_class]" = "20" ) )) END'
    assert(output(s) == exp)


def test_and_expressions():
    s = '''
    CLASS
        EXPRESSION ("[style_class]" = "10" AND "[style_class]" = "20")
    END
    '''

    exp = 'CLASS EXPRESSION (( ( "[style_class]" = "10" ) and ( "[style_class]" = "20" ) )) END'
    assert(output(s) == exp)

    s = '''
    CLASS
        EXPRESSION ("[style_class]" = "10" && "[style_class]" = "20")
    END
    '''

    exp = 'CLASS EXPRESSION (( ( "[style_class]" = "10" ) and ( "[style_class]" = "20" ) )) END'
    assert(output(s) == exp)


def test_not_expressions():
    s = '''
    CLASS
        EXPRESSION NOT("[style_class]" = "20")
    END
    '''

    exp = 'CLASS EXPRESSION not (( "[style_class]" = "20" )) END'
    assert(output(s) == exp)

    s = '''
    CLASS
        EXPRESSION !("[style_class]" = "20")
    END
    '''

    exp = 'CLASS EXPRESSION not (( "[style_class]" = "20" )) END'
    assert(output(s) == exp)


def test_processing_directive():

    s = """
    LAYER
        NAME 'ProcessingLayer'
        PROCESSING 'BANDS=1'
        PROCESSING 'CONTOUR_ITEM=elevation'
        PROCESSING 'CONTOUR_INTERVAL=20'
    END
    """

    #print(output(s))
    exp = "LAYER NAME 'ProcessingLayer' PROCESSING 'BANDS=1' PROCESSING 'CONTOUR_ITEM=elevation' PROCESSING 'CONTOUR_INTERVAL=20' END"
    assert(output(s) == exp)


def test_config_directive():

    s = """
    MAP
        NAME 'ConfigMap'
        CONFIG MS_ERRORFILE 'stderr'
        CONFIG 'PROJ_DEBUG' 'OFF'
        CONFIG 'ON_MISSING_DATA' 'IGNORE'
    END
    """

    exp = "MAP NAME 'ConfigMap' CONFIG MS_ERRORFILE 'stderr' CONFIG 'PROJ_DEBUG' 'OFF' CONFIG 'ON_MISSING_DATA' 'IGNORE' END"
    assert(output(s) == exp)

def test_multiple_composites():
    """
    Allow for multiple root composites
    This allows for easier addition of CLASSES and STYLES
    """

    s = """
    CLASS
        Name "Name1"
    END
    CLASS
        Name "Name2"
    END
    """
    exp = "CLASS NAME 'Name1' END CLASS NAME 'Name2' END"
    assert(output(s) == exp)

def test_map():
    s = """
    MAP
        LAYER
            NAME 'test'
        END
    END
    """

    exp = "MAP LAYER NAME 'test' END END"
    assert(output(s) == exp)


def test_oneline_composites():
    """
    Test a composite on one line is parsed correctly
    """

    s = """
    CLASS
    PROJECTION "init=epsg:2056"
    END
    END
    """

    # put on one line
    exp = ' '.join(s.split())
    assert(output(s) == exp)

def test_querymap():

    s = """
    MAP
        QUERYMAP
           COLOR 255 255 0
           SIZE -1 -1
           STATUS OFF
           STYLE HILITE
         END
    END
    """
    #print output(s)
    exp = "MAP QUERYMAP COLOR 255 255 0 SIZE -1 -1 STATUS OFF STYLE HILITE END END"
    assert(output(s) == exp)

def test_output_format():

    s = """
    OUTPUTFORMAT
        NAME "shapezip"
        DRIVER "OGR/ESRI Shapefile"
        TRANSPARENT FALSE
        IMAGEMODE FEATURE
    END
    """
    exp = "OUTPUTFORMAT NAME 'shapezip' DRIVER 'OGR/ESRI Shapefile' TRANSPARENT FALSE IMAGEMODE FEATURE END"
    assert(output(s) == exp)

def test_auto_projection():
    """
    Test an AUTO projection
    """

    s = """
    MAP
        PROJECTION
            AUTO
        END
    END
    """
    exp = "MAP PROJECTION AUTO END END"
    #print(output(s))
    assert(output(s) == exp)

def test_runtime_expression():
    s = """
    CLASS
      EXPRESSION ( [EPPL_Q100_] = %eppl% )		   
    END
    """
    exp = "CLASS EXPRESSION (( [EPPL_Q100_] = %eppl% )) END"
    #print(output(s))
    assert(output(s) == exp)

def test_mutliple_output_formats():
    """
    https://github.com/geographika/mappyfile/issues/20
    """
    s = """
    OUTPUTFORMAT
        FORMATOPTION "FORM=zip"
        FORMATOPTION "SPATIAL_INDEX=YES"
    END
    """
    exp = "OUTPUTFORMAT FORMATOPTION 'FORM=zip' FORMATOPTION 'SPATIAL_INDEX=YES' END"
    assert(output(s) == exp)

def test_config_case():
    """
    https://github.com/geographika/mappyfile/issues/18
    """

    s = """
    MAP
	    CONFIG "PROJ_LIB" "projections"
    END
    """
    exp = "MAP CONFIG 'PROJ_LIB' 'projections' END"
    assert(output(s) == exp)

def test_ne_comparison():
    """
    IS NOT is not valid
    NE (Not Equals) should be used instead 
    """
    s = """
    CLASS
        # EXPRESSION ( "[building]" IS NOT NULL) # incorrect syntax
        EXPRESSION ( "[building]" NE NULL)
    END
    """
    exp = 'CLASS EXPRESSION (( "[building]" NE NULL )) END'
    assert(output(s) == exp)

def test_eq_comparison():
    """
    Case is not changed for comparison (EQ/eq stay the same)
    """
    s = """
    CLASS
        EXPRESSION ( "[building]" eq NULL)
    END
    """
    exp = 'CLASS EXPRESSION (( "[building]" eq NULL )) END'
    #print(output(s))
    assert(output(s) == exp)

def test_no_linebreaks():
    """
    Check that classes can be nested on a single line
    """
    s = """CLASS NAME 'Test' STYLE OUTLINECOLOR 0 0 0 END END"""
    exp = "CLASS NAME 'Test' STYLE OUTLINECOLOR 0 0 0 END END"
    assert(output(s) == exp)

def run_tests():        
    #pytest.main(["tests/test_snippets.py::test_style_pattern"])
    pytest.main(["tests/test_snippets.py"])

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    #test_no_linebreaks()
    run_tests()
    print("Done!")
